import inspect
import matplotlib.dates
import matplotlib.ticker
import sys
import datetime
import numpy as np
import calendar

import verif.util


def get_all():
    """
    Returns a dictionary of all axis classes where the key is the class
    name (string) and the value is the class object
    """
    temp = inspect.getmembers(sys.modules[__name__], inspect.isclass)
    return temp


def get(name):
    """ Returns an instance of an object with the given class name """
    axes = get_all()
    a = None
    for axis in axes:
        if name == axis[0].lower():
            a = axis[1]()
    if a is None:
        verif.util.error("No axis by name '%s'" % name)
    return a


def get_time_axes():
    return [axis[1]() for axis in get_all() if hasattr(axis[1], "compute_from_times")]


def get_leadtime_axes():
    return [axis[1]() for axis in get_all() if hasattr(axis[1], "compute_from_leadtimes")]
    return [verif.axis.Leadtime(), verif.axis.Leadtimeday()]

def get_spatial_axes():
    return [axis[1]() for axis in get_all() if axis[1]().is_location_like]


class Axis(object):
    """ Represents axes that data can be arranged by

    Attributes:
       is_continuous (bool): Does this axis represent a notion that is
          continuous (does it make sense to connect dots with a line on a
          graph)? For example, Location is not a continuous axis.
       is_location_like (bool): Does this axis related to the notion of location?
       is_time_like (bool): Does this axis have anything to do with the notion of time?
       formatter (matplotlib.ticker.Formatter): What format should this axis
          have? Returns an mpl Formatter
    """
    is_continuous = True
    is_location_like = False
    is_time_like = False
    fmt = "%f"

    @classmethod
    def name(cls):
        name = cls.__name__
        return name

    def label(self, variable):
        """ Returns an appropriate string for labeling an axis on a plot """
        raise NotImplementedError()

    def formatter(self, variable):
        """ How should ticks be generated for this axis? """
        return matplotlib.ticker.ScalarFormatter()

    def __eq__(self, other):
        return self.__class__ == other.__class__

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        # TODO
        return 1


class Time(Axis):
    """ Forecast initialization time """
    is_time_like = True
    fmt = "%Y-%m-%d %H:%M:%S"

    def formatter(self, variable):
        return matplotlib.dates.DateFormatter('\n%Y-%m-%d')

    def label(self, variable):
        return "Date"


class Leadtime(Axis):
    """ Forecast lead-time """

    def label(self, variable):
        return "Lead time (h)"

    def compute_from_leadtimes(self, leadtimes):
        return leadtimes


class Leadtimeday(Axis):
    """ Forecast lead-time aggregated over a day"""

    def label(self, variable):
        return "Lead time (day)"

    def compute_from_leadtimes(self, leadtimes):
        return np.array([int(d/24) for d in leadtimes])


class Location(Axis):
    is_continuous = False
    is_location_like = True

    def label(self, variable):
        return "Location"


class Lat(Axis):
    is_continuous = False
    is_location_like = True

    def label(self, variable):
        return "Latitude ($^o$)"


class Lon(Axis):
    is_continuous = False
    is_location_like = True

    def label(self, variable):
        return "Longitude ($^o$)"


class Elev(Axis):
    is_continuous = False
    is_location_like = True

    def label(self, variable):
        return "Elevation (m)"


class All(Axis):
    """ No aggregation done """
    is_continuous = False


class No(Axis):
    """ Entire 3D array is serialized """
    is_continuous = False


class Year(Axis):
    is_time_like = True
    fmt = "%Y"

    def formatter(self, variable):
        return matplotlib.dates.DateFormatter('\n%Y')

    def label(self, variable):
        return "Year"

    def compute_from_times(self, times):
        dts = [datetime.datetime.utcfromtimestamp(i) for i in times]
        dts = [d.replace(month=1, day=1, hour=0, minute=0, second=0) for d in dts]
        return np.array([calendar.timegm(dt.timetuple()) for dt in dts])


class Month(Axis):
    is_time_like = True
    fmt = "%Y/%m"

    def formatter(self, variable):
        return matplotlib.dates.DateFormatter('%b\n%Y')

    def label(self, variable):
        return "Month"

    def compute_from_times(self, times):
        dts = [datetime.datetime.utcfromtimestamp(i) for i in times]
        dts = [d.replace(day=1, hour=0, minute=0, second=0) for d in dts]
        return np.array([calendar.timegm(dt.timetuple()) for dt in dts])


class Week(Axis):
    is_time_like = True
    fmt = "%Y/%U"

    def formatter(self, variable):
        return matplotlib.dates.DateFormatter('%Y-%U')

    def label(self, variable):
        return "Week"

    def compute_from_times(self, times):
        dts = [datetime.datetime.utcfromtimestamp(i) for i in times]
        dts = [d.replace(hour=0, minute=0, second=0) for d in dts]
        # Reset datetime such that it is for the first day of the week
        # That is subtract the day of the week from the date
        dts = [d - datetime.timedelta(days=d.weekday()) for d in dts]
        return np.array([calendar.timegm(dt.timetuple()) for dt in dts])


class Timeofday(Axis):
    def label(self, variable):
        return "Time of day"

    def compute_from_times(self, times):
        return (times % 86400) / 3600


class Dayofyear(Axis):
    def label(self, variable):
        return "Day of year"

    def compute_from_times(self, times):
        dts = [datetime.datetime.utcfromtimestamp(i) for i in times]
        dts = [d.replace(year=2000) for d in dts]
        return np.array([(x - datetime.datetime(year=2000, month=1, day=1)).days + 1 for x in dts])


class Day(Axis):
    is_time_like = True
    fmt = "%Y/%m/%d"

    def formatter(self, variable):
        return matplotlib.dates.DateFormatter('\n%Y-%m-%d')

    def label(self, variable):
        return "Day"

    def compute_from_times(self, times):
        dts = [datetime.datetime.utcfromtimestamp(i) for i in times]
        dts = [d.replace(hour=0, minute=0, second=0) for d in dts]
        return np.array([calendar.timegm(dt.timetuple()) for dt in dts])


class Dayofmonth(Axis):
    def label(self, variable):
        return "Day of month"

    def compute_from_times(self, times):
        return np.array([datetime.datetime.utcfromtimestamp(i).day for i in times])


class Monthofyear(Axis):
    def label(self, variable):
        return "Month of year"

    def compute_from_times(self, times):
        return np.array([datetime.datetime.utcfromtimestamp(i).month for i in times])


class Obs(Axis):
    def formatter(self, variable):
        return variable.formatter

    def label(self, variable):
        return "Observed " + variable.name + " (" + variable.units + ")"


class Fcst(Axis):
    def formatter(self, variable):
        return variable.formatter

    def label(self, variable):
        return "Forecasted " + variable.name + " (" + variable.units + ")"


class Threshold(Axis):
    def formatter(self, variable):
        return variable.formatter

    def label(self, variable):
        return variable.name + " (" + variable.units + ")"



import sys
import numpy as np


class Interval(object):
    """ Represents mathematical intervals in the real number line

    The following forms are supported, where [,] are inclusive and (,) are not:
    [lower, upper], (lower, upper], [lower, upper), or (lower, upper)

    Attributes:
       lower (float): The lower boundary of the interval
       upper (float): The upper boundary of the interval
       lower_eq (bool): True to include the lower boundary in the interval
       upper_eq (bool): True to include the upper boundary in the interval
    """
    def __init__(self, lower, upper, lower_eq, upper_eq):
        self.lower = lower if lower is not None else -np.inf
        self.upper = upper if upper is not None else np.inf
        self.lower_eq = lower_eq
        self.upper_eq = upper_eq

    def within(self, x):
        """ Is one or more values within the interval?

        Args:
           x (float or np.array): value(s)

        Returns:
           bool or np.array(bool): True if the value(s) is in the interval, False otherwise.
           If any of the inputs have np.nan values, then a masked array is
           returned with the corresponding answers masked out.
        """
        if isinstance(x, np.ndarray):
            I = np.where(np.isnan(x) == 0)
            """
            Only evaluate valid values, in order to avoid a warning. Then fill
            these values into a masked array
            """
            values = np.zeros(x.shape, bool)
            is_above = (x[I] > self.lower) | (self.lower_eq and x[I] == self.lower)
            is_below = (x[I] < self.upper) | (self.upper_eq and x[I] == self.upper)
            values[I] = is_above & is_below
            values = np.ma.masked_array(values, mask=np.isnan(x))
        else:
            if np.isnan(x):
                values = np.nan
            else:
                is_above = (x > self.lower) | (self.lower_eq and x == self.lower)
                is_below = (x < self.upper) | (self.upper_eq and x == self.upper)
                values = (is_above & is_below)

        return values

    @property
    def center(self):
        """ float: The center of the interval """
        if self.lower == -np.inf and self.upper == np.inf:
            return 0
        elif self.lower == -np.inf:
            return self.upper
        elif self.upper == np.inf:
            return self.lower
        else:
            return (self.lower + self.upper)/2

    def __str__(self):
        lower_bracket = "("
        if self.lower_eq:
            lower_bracket = "["
        upper_bracket = ")"
        if self.upper_eq:
            upper_bracket = "]"
        return "%s%f, %f%s" % (lower_bracket, self.lower, self.upper, upper_bracket)

    def __eq__(self, other):
        return self.lower == other.lower and\
               self.upper == other.upper and\
               self.lower_eq == other.lower_eq and\
               self.upper_eq == other.upper_eq

    def __ne__(self, other):
        return not self.__eq__(other)



import sys
import inspect


class MetricType(object):
    def __eq__(self, other):
        return self.__class__ == other.__class__

    def __ne__(self, other):
        return not self.__eq__(other)

    @classmethod
    def name(cls):
        name = cls.__name__
        return name


class Deterministic(MetricType):
    description = "Deterministic"
    pass


class Probabilistic(MetricType):
    description = "Probabilistic"
    pass


class Threshold(MetricType):
    description = "Threshold"
    pass


class Diagram(MetricType):
    description = "Special diagrams"
    pass



import inspect
import numpy as np
import sys

import verif.util


def get_all():
    """ Returns a list of all aggregator classes """
    temp = inspect.getmembers(sys.modules[__name__], inspect.isclass)
    return [i[1] for i in temp if i[0] != "Aggregator"]


def get(name):
    """ Returns an instance of an object with the given class name

    Arguments:
       name (str): The name of the class. Use a number between 0 and 1 to get the
          corresponding quantile aggregator
    """
    aggregators = get_all()
    a = None
    for aggregator in aggregators:
        if name == aggregator.name():
            a = aggregator()
    if a is None and verif.util.is_number(name):
        a = Quantile(float(name))

    if a is None:
        verif.util.error("No aggregator by the name '%s'" % name)

    return a


class Aggregator(object):
    """ Base class for aggregating an array (computing a scalar value from an array)

    Usage:
       mean = verif.aggregator.Mean()
       mean(np.array([1,2,3]))

    Attributes:
       name: A string representing the name of the aggregator
    """

    def __call__(self, array, axis=None):
        """ Compute the aggregated value. Returns a scalar value.

        Arguments:
           array (np.array): A 1D numpy array
           axis (int): Axis to aggregate across
        """
        raise NotImplementedError()

    @classmethod
    def name(cls):
        return cls.__name__.lower()

    def __hash__(self):
        # TODO
        return 1

    def __eq__(self, other):
        return self.__class__ == other.__class__

    def __ne__(self, other):
        return not self.__eq__(other)


class Mean(Aggregator):
    def __call__(self, array, axis=None):
        return np.mean(array, axis=axis)


class Median(Aggregator):
    def __call__(self, array, axis=None):
        return np.median(array, axis=axis)


class Min(Aggregator):
    def __call__(self, array, axis=None):
        return np.min(array, axis=axis)


class Max(Aggregator):
    def __call__(self, array, axis=None):
        return np.max(array, axis=axis)


class Std(Aggregator):
    def __call__(self, array, axis=None):
        return np.std(array, axis=axis)


class Variance(Aggregator):
    def __call__(self, array, axis=None):
        return np.var(array, axis=axis)


class Iqr(Aggregator):
    def __call__(self, array, axis=None):
        return np.percentile(array, 75, axis=axis) - np.percentile(array, 25, axis=axis)


class Range(Aggregator):
    def __call__(self, array, axis=None):
        return verif.util.nprange(array, axis=axis)


class Count(Aggregator):
    def __call__(self, array, axis=None):
        return verif.util.numvalid(array, axis=axis)


class Sum(Aggregator):
    def __call__(self, array, axis=None):
        return np.sum(array, axis=axis)


class Meanabs(Aggregator):
    """ The mean of the absolute values of the array """
    def __call__(self, array, axis=None):
        return np.mean(np.abs(array), axis=axis)


class Absmean(Aggregator):
    """ Absolute value of the mean of the array """
    def __call__(self, array, axis=None):
        return np.abs(np.mean(array, axis=axis))


class Quantile(Aggregator):
    def __init__(self, quantile):
        """ Returns a certain quantile from the array

        Arguments:
           quantile (float): A value between 0 and 1 inclusive
        """
        self.quantile = quantile
        if self.quantile < 0 or self.quantile > 1:
            verif.util.error("Quantile must be between 0 and 1")

    def __call__(self, array, axis=None):
        return np.percentile(array, self.quantile*100, axis=axis)

    def __eq__(self, other):
        return (self.__class__ == other.__class__) and (self.quantile == other.quantile)



from __future__ import print_function
import matplotlib.dates
import calendar
import copy
import datetime
import matplotlib.pyplot as mpl
import numpy as np
import os
import re
import sys
import textwrap
import netCDF4

import verif.interval

"""
There are 4 ways to represent time in verif:
   1) unixtime (seconds since 1970). Used internally in verif.data and also
      optionally in input files.
   2) date (YYYYYMMDD). Used on the command-line to specify dates and also
      optionally in input files.
   3) datenum (Some kind of matplotlib way to represent date as integer). Times
      must be in this format to make it easier to put automatic ticks on time axes
   4) datetime (a full date object). Not really used in verif, except when a
      formatted datestring is needed.

Below are functions that convert between the first 3
"""


def date_to_datenum(date):
    """ Converts date in YYYYMMDD format into datenum

    Arguments:
       date (int): date in YYYYMMDD format

    Returns:
       int: datenum value
    """
    year = int(date / 10000)
    month = int(date / 100 % 100)
    day = int(date % 100)
    return matplotlib.dates.date2num(datetime.datetime(year, month, day, 0))


def unixtime_to_datenum(time):
    """ Converts unixtime into datenum

    Arguments:
       time (int): unixtime in seconds since 1970

    Returns:
       int: datenum value
    """
    dt = datetime.datetime.utcfromtimestamp(time)
    return matplotlib.dates.date2num(dt)


def bin(x, y, edges, func=np.nanmean):
    """ Bin the x-values using edges and return the average values

    Arguments:
       x (np.array): x-axis values
       y (np.array): y-axis values
       edges (np.array): bins with these edges
       func: what function to use when binning


    Returns:
       xx (np.array): average values of x in each bin. Length is one less than
          edges
       yy (np.array): average values of y in each bin
    """
    xx = np.nan*np.zeros(len(edges)-1)
    yy = np.nan*np.zeros(len(edges)-1)

    for i in range(len(edges)-1):
        I = np.where((x >= edges[i]) & (x < edges[i+1]))[0]
        if len(I) > 0:
            xx[i] = np.nanmean(x[I])
            yy[i] = func(y[I])

    return xx, yy


def datenum_to_date(datenum):
    """ Converts datenum into YYYYMMDD value

    Arguments:
       date (int): datenum

    Returns:
       int: date in YYYYMMDD
    """
    return int(matplotlib.dates.num2date(datenum).strftime("%Y%m%d"))


def date_to_unixtime(date):
    """ Convert YYYYMMDD to unixtime

    Arguments:
       date (int): YYYYMMDD

    Returns:
       int: unixtime
    """
    year = date // 10000
    month = date // 100 % 100
    day = date % 100
    ut = calendar.timegm(datetime.datetime(year, month, day).timetuple())
    return ut


def date_to_unixtime_slow(date):
    ut = calendar.timegm(datetime.datetime.strptime(str(date), "%Y%m%d").timetuple())
    return ut


def unixtime_to_date(unixtime):
    """ Convert unixtime to YYYYMMDD

    Arguments:
       unixtime (int): unixtime

    Returns:
       int: date in YYYYMMDD
    """
    dt = datetime.datetime.utcfromtimestamp(int(unixtime))
    date = dt.year * 10000 + dt.month * 100 + dt.day
    return date


def red(text):
    """ Print text in red to the console """
    return "\033[31m" + text + "\033[0m"


def green(text):
    """ Print text in green to the console """
    return "\033[32m" + text + "\033[0m"


def yellow(text):
    """ Print text in yellow to the console """
    return "\033[33m" + text + "\033[0m"


def remove_margin():
    mpl.subplots_adjust(left=0, right=1, bottom=0, top=1, wspace=0, hspace=0)


def experimental():
    return yellow("(experimental)")


def error(message):
    """ Write error message to console and abort """
    print("\033[1;31mError: " + message + "\033[0m")
    sys.exit(1)


def warning(message):
    """ Write a warning message to console """
    print("\033[1;33mWarning: " + message + "\033[0m")


def parse_numbers(numbers, is_date=False):
    """
    Parses numbers from an input string. Recognizes MATLAB syntax, such as:
    3              single numbers
    3,4,5          list of numbers
    3:5            number range
    3:2:12         number range with a step size of 2
    3,4:6,2:5:9,6  combinations

    Aborts if the number cannot be parsed. Expect round-off errors for values
    below about 1e-4.

    Arguments:
       numbers (str): String of numbers
       is_date (bool): True if values should be interpreted as YYYYMMDD

    Returns:
       list: parsed numbers
    """
    # Check if valid string
    if any(char not in set('-01234567890.:,') for char in numbers):
        error("Could not translate '" + numbers + "' into numbers")

    values = list()
    commaLists = numbers.split(',')
    for commaList in commaLists:
        colonList = commaList.split(':')
        for w in colonList:
            if w == "":
                verif.util.error("Could not parse '%s'. Empty value." % (numbers))
        if len(colonList) == 1:
            values.append(float(colonList[0]))
        elif len(colonList) <= 3:
            start = float(colonList[0])
            step = 1
            if len(colonList) == 3:
                step = float(colonList[1])
            if step == 0:
                verif.util.error("Could not parse '%s': Step cannot be 0." % (numbers))
            stepSign = step / abs(step)
            # arange does not include the end point:
            end = float(colonList[-1]) + stepSign * 0.0001
            if is_date:
                date = min(start, end)
                curr = list()
                while date <= max(start, end):
                    curr.append(date)
                    date = get_date(date, step)
                values = values + list(curr)
            else:
                # Note: Values are rounded, to avoid problems with floating point
                # comparison for strings like 0.1:0.1:0.9
                values = values + list(np.round(np.arange(start, end, step), 7))
        else:
            error("Could not translate '" + numbers + "' into numbers")
        if is_date:
            for i in range(0, len(values)):
                values[i] = int(values[i])
    return values


def parse_ints(numbers):
    return [int(num) for num in parse_numbers(numbers)]


def parse_dates(numbers):
    return parse_numbers(numbers, True)


def parse_colors(colors):
    """
    """
    firstList = colors.split(",")
    numList = []
    finalList = []

    for string in firstList:
        if "[" in string:   # for rgba args
            if not numList:
                string = string.replace("[", "")
                numList.append(float(string))
            else:
                verif.util.error("Invalid rgba arg \"{}\"".format(string))

        elif "]" in string:
            if numList:
                string = string.replace("]", "")
                numList.append(float(string))
                finalList.append(numList)
                numList = []
            else:
                verif.util.error("Invalid rgba arg \"{}\"".format(string))

        # append to rgba lists if present, otherwise grayscale intensity
        elif verif.util.is_number(string):
            if numList:
                numList.append(float(string))
            else:
                finalList.append(string)

        else:
            if not numList:  # string args and hexcodes
                finalList.append(string)
            else:
                verif.util.error("Cannot read color args.")
    return finalList


def subplot(i, N):
    """ Sets up subplot for index i (starts at 0) out of N """
    [X, Y] = get_subplot_size(N)
    mpl.subplot(Y, X, i + 1)


def get_subplot_size(N):
    Y = 1
    if N > 4:
        Y = np.ceil(np.sqrt(N) / 1.5)
    X = np.ceil(N / Y)
    return [int(X), int(Y)]


def get_map_resolution(lats, lons):
    dlat = (max(lats) - min(lats))
    dlon = (max(lons) - min(lons))
    scale = max(dlat, dlon)
    if np.isnan(scale):
        res = "c"
    elif scale > 60:
        res = "c"
    elif scale > 1:
        res = "i"
    elif scale > 0.001:
        res = "h"
    elif scale > 0.0001:
        res = "f"
    else:
        res = "c"
    return res


def get_cartopy_map_resolution(lats, lons):
    dlat = (max(lats) - min(lats))
    dlon = (max(lons) - min(lons))
    scale = max(dlat, dlon)
    return 5
    # TODO: Implement this
    if np.isnan(scale):
        res = "c"
    elif scale > 60:
        res = "c"
    elif scale > 1:
        res = "i"
    elif scale > 0.001:
        res = "h"
    elif scale > 0.0001:
        res = "f"
    else:
        res = "c"
    return res


def fill(x, y_lower, y_upper, col, alpha=1, zorder=0, hatch=''):
    """ Fill an area between two curves

    Fill an area along x, between y_lower and y_upper. Both y_lower and y_upper most
    correspond to points in x (i.e. be in the same order)

    Arguments:
       x (np.array): x-axis values
       y_lower (np.array): y-axis values for lower envelope
       y_upper (np.array): y-axis values for upper envelope
       col: Color of filled area in any format understood by mpl.fill
       alpha: alpha of filled area
       zorder: zorder
       hatch: any hatch string understood by mpl.fill
    """
    # Populate a list of non-missing points
    X = list()
    Y = list()
    for i in range(0, len(x)):
        if not(np.isnan(x[i]) or np.isnan(y_lower[i])):
            X.append(x[i])
            Y.append(y_lower[i])
    for i in range(len(x) - 1, -1, -1):
        if not (np.isnan(x[i]) or np.isnan(y_upper[i])):
            X.append(x[i])
            Y.append(y_upper[i])
    if len(X) > 0:
        mpl.fill(X, Y, facecolor=col, alpha=alpha, linewidth=0, zorder=zorder,
              hatch=hatch)


def clean(data):
    """ Copy and sanitize data from a netCDF4 variable

    Arguments:
       data: A netCDF4 variable

    Returns:
       np.array: A numpy array where invalid values have been set to np.nan

    """
    if len(data.shape) == 1 and data.shape[0] == 0:
        return np.zeros(0)

    data = data[:].astype(float)
    q = np.ma.filled(data, fill_value=-999)
    # Remove missing values. Convert to -999 and then back to nan to avoid
    # warning messages when doing <, >, and == comparisons with nan.
    q[np.isnan(q)] = -999
    q[(q == -999) | (q > 1e30)] = np.nan
    return q


def get_date(date, diff):
    """ Date calculation: Adds 'diff' to 'date'

    Arguments:
       date (int): An integer of the form YYYYMMDD
       diff (int): Number of days to add to date

    Returns:
       int: A new date in the form YYYYMMDD
    """
    year = int(date / 10000)
    month = int(date / 100 % 100)
    day = int(date % 100)
    date2 = datetime.datetime(year, month, day, 0) + datetime.timedelta(diff)
    return int(date2.strftime('%Y%m%d'))


def nanmean(data, **args):
    return np.ma.filled(np.ma.masked_array(data, np.isnan(data)).mean(**args),
          fill_value=np.nan)


def nanmedian(data, **args):
    I = np.where(np.isnan(data.flatten()) == 0)[0]
    return np.median(data.flatten()[I])


def nanmin(data, **args):
    return np.ma.filled(np.ma.masked_array(data, np.isnan(data)).min(**args),
          fill_value=np.nan)


def nanmax(data, **args):
    return np.ma.filled(np.ma.masked_array(data, np.isnan(data)).max(**args),
          fill_value=np.nan)


def nanstd(data, **args):
    return np.ma.filled(np.ma.masked_array(data, np.isnan(data)).std(**args),
          fill_value=np.nan)


def nanpercentile(data, pers):
    I = np.where(np.isnan(data.flatten()) == 0)[0]
    p = np.nan
    if len(I) > 0:
        p = np.percentile(data.flatten()[I], pers)
    return p


def numvalid(data, axis=None):
    return np.sum(np.isnan(data) == 0, axis=axis)


def nprange(data, axis=None):
    return np.max(data, axis=axis) - np.min(data, axis=axis)


def intersect(list1, list2):
    return list(set(list1) & set(list2))


def get_screen_width():
    """ How many character wide is the console? """
    rows, columns = os.popen('stty size', 'r').read().split()
    columns = int(columns)
    return columns


def is_number(s):
    """ Returns true if x is a scalar number """
    try:
        float(s)
        return True
    except ValueError:
        return False


def distance(lat1, lon1, lat2, lon2):
    """
    Computes the great circle distance between two points using the
    haversine formula. Values can be vectors.
    """
    # Convert from degrees to radians
    pi = 3.14159265
    lon1 = lon1 * 2 * pi / 360
    lat1 = lat1 * 2 * pi / 360
    lon2 = lon2 * 2 * pi / 360
    lat2 = lat2 * 2 * pi / 360
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = np.sin(dlat / 2)**2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2)**2
    c = 2 * np.arcsin(np.sqrt(a))
    distance = 6.367e6 * c
    return distance


def apply_threshold(array, bin_type, threshold, upper_threshold=None):
    """ Use bin_type to turn array into binary values """
    I = np.isnan(array) == 0
    array = copy.deepcopy(array)
    if bin_type == "below":
        array[I] = array[I] < threshold
    elif bin_type == "below=":
        array[I] = array[I] <= threshold
    elif bin_type == "above":
        array[I] = array[I] > threshold
    elif bin_type == "above=":
        array[I] = array[I] >= threshold
    elif upper_threshold is None:
        error("Cannot apply thresholding with bin_type '%s'" % bin_type)
    elif bin_type == "within":
        array[I] = (array[I] > threshold) & (array[I] < upper_threshold)
    elif bin_type == "within=":
        array[I] = (array[I] > threshold) & (array[I] <= upper_threshold)
    elif bin_type == "=within":
        array[I] = (array[I] >= threshold) & (array[I] < upper_threshold)
    elif bin_type == "=within=":
        array[I] = (array[I] >= threshold) & (array[I] <= upper_threshold)
    array = np.array(array, float)
    return array


def apply_threshold_prob(array, bin_type, upper_array=None):
    """
    Compute the probability of x given bin_type

    array          The CDF at the threshold
    bin_type
    upper_array    The CDF of the upper threshold (only needed with
                   bin_type = *within*)
    """
    if bin_type == "below":
        array = array
    elif bin_type == "below=":
        array = array
    elif bin_type == "above":
        array = 1 - array
    elif bin_type == "above=":
        array = 1 - array
    elif upper_array is None:
        error("Cannot apply thresholding with bin_type '%s'" % bin_type)
    elif re.compile("within").match(bin_type):
        array = upper_array - array
    return array


def get_threshold_string(bin_type):
    s = ""
    if bin_type == "below":
        s = "<"
    elif bin_type == "below=":
        s = "<="
    elif bin_type == "above":
        s = ">"
    elif bin_type == "above=":
        s = ">="
    else:
        error("Cannot get threshold string for bin_type '%s'" % bin_type)
    return s


def deg2rad(deg):
    """ Convert degrees to radians """
    return deg * np.pi / 180.0


def get_square_axis_limits(xlim, ylim):
    axismin = min(min(ylim), min(xlim))
    axismax = max(max(ylim), max(xlim))
    return [axismin, axismax]


def get_intervals(bin_type, thresholds):
    """ Returns a list of interval objects. If bin_type is *within*, then
    intervals are formed by using each pair of consecutive thresholds. For
    bin_type below* the interval [-np.inf, threshold] is used and for bin_type
    above* the inveral [threshold, np.inf] is used.

    Arguments:
    bin_type       one of below, below=, within, =within, within=, =within=,
                   above, above=
    thresholds     numy array of thresholds
    """
    if thresholds is None:
        return [verif.interval.Interval(-np.inf, np.inf, True, True)]

    intervals = list()
    N = len(thresholds)
    if re.compile(".*within.*").match(bin_type):
        N = N - 1
    for i in range(0, N):
        lower_eq = False
        upper_eq = False

        if bin_type in ["below", "below="]:
            lower = -np.inf
            upper = thresholds[i]
        elif bin_type in ["above", "above="]:
            lower = thresholds[i]
            upper = np.inf
        elif bin_type in ["within", "=within", "within=", "=within="]:
            lower = thresholds[i]
            upper = thresholds[i+1]
        else:
            verif.util.error("Unrecognized bintype")
        if bin_type in ["below=", "within=", "=within="]:
            upper_eq = True
        if bin_type in ["above=", "=within", "=within="]:
            lower_eq = True
        intervals.append(verif.interval.Interval(lower, upper, lower_eq, upper_eq))
    return intervals


def almost_equal(value1, value2, tol=1e-7):
    return abs(value1 - value2) < tol


def is_valid_nc(filename):
    """ Return True if the file is a valid NetCDF file """
    try:
        file = netCDF4.Dataset(filename, 'r')
        file.close()
        return True
    except Exception:
        return False


def get_distance_matrix(locations):
    """ Computes distance matrix betweeen locations

    Arguments:
       locations (list): List of verif.location.Location)

    Returns:
       np.array: 2D array with the distance between each pair of locations in meters
    """

    N = len(locations)
    dist = np.zeros([N, N])
    for i in range(N):
        for j in range(N):
            dist[i, j] = locations[i].get_distance(locations[j])
    return dist


def proj4_string_to_dict(string):
    """ Parse a proj4 string and create a dictionary

    Arguments:
       string (str): A proj4 string like "+proj=lcc +lat_0=63
       +lon_0=15 +lat_1=63 +lat_2=63 +no_defs +R=6.371e+06"
    Returns:
       dict: A dictionary of attributes and values, e.g =proj=
    """
    r = dict()
    pairs = string.split(' ')
    for pair in pairs:
        keyvalue = pair.split('=')
        if len(keyvalue) > 2:
            verif.util.error("Could not parse proj4 parameter: %s" % pair)
        key = keyvalue[0]
        if len(keyvalue) == 1:
            value = True
        else:
            value = keyvalue[1]
        # Try converting to number
        try:
            value = float(value)
        except Exception:
            pass
        r[key] = value
    return r


def parse_label(string):
    """
    Replaces newlines
    """
    return str(string).replace('\\n', '\n')
